#include "function.h"
#include "compile.h"
#include "colon-stack.h"
#include "colon-branch.h"
#include "stack.h"
#include "memory.h"
#include "colon-state.h"
#include "colon-compiler.h"
#include "colon-constants.h"

/*
if       \ forward conditional branch 
ahead    \ forward unconditional branch 
then     \ resolve forward branch

begin    \ mark backward destination 
until    \ resolve backward with conditional branch 
again    \ resolve backward with unconditional branch 

cs-pick  \ copy item on control flow stack 
cs-roll  \ reorder items on control flow stack 
         \ just use the pointer stack versions! 
*/

C( P: -- ftok )
void VVcompile_if(void)
{
  op_t *op;
  reg_t reg;
  varbl *var;
  arg_array *reg_state;
  
  ppush(&dstack); VVget_stack(); VVstack_pop(); var = ppop();
  ppush(var); VVvar_reg(); reg = dpop();

  /* now I need some way to save the state etc. */
  /* just before this jump */

  VVreg_state(); reg_state = ppop();   /* get the register state */

  op = allocate(sizeof(op_t));
  op->opcode = GOTOQ;
  op->r1 = reg;
  op->label = 0;     /* don't resolve this now */
  op->value =(uval)reg_state;  /* use this as scratch for the register state */
  ppush(op);
  VVappend_def();

  ppush(var);  VVrelease_var();

  ppush(op);
}

C( P: -- ftok )
void VVcompile_ahead(void)
{
  op_t *op;
  arg_array *reg_state;
  /* now I need some way to save the state etc. */
  /* just before this jump */

  VVreg_state(); reg_state = ppop();
  
  op = allocate(sizeof(op_t));
  op->opcode = GOTOQ;
  op->r1 = ZERO;
  op->label = 0;     /* don't resolve this now */
  op->value =(uval)reg_state;  /* use this as scratch for the register state */
  ppush(op);
  VVappend_def();

  ppush(op);
}

C( P: ftok -- )
void VVcompile_then(void)
{
  op_t *op1, *op2;
  op1 = ppop();

  ppush((ref)op1->value); VVload_arg_regs();
  ppush((ref)op1->value); VVunlock_arg_regs();

  op2 = allocate(sizeof(op_t));
  op2->opcode = NOP;
  ppush(op2);
  VVappend_def();

  op1->label = op2;
}

#define MY_NUM 342
C( P: -- btok )
void VVcompile_begin(void)
{
  op_t * op;
  arg_array *reg_state;

  /* if I don't have at least one free variable here, and this is matched by an */
  /* until I'm in trouble.  For the moment I don't know how to guarantee a free */
  /* variable.  load a dummy constant maybe.  */

  dpush(MY_NUM); VVget_dummy(); VVvar_reg(); dpop();  /* assign a dummy to a register */
  VVreg_state(); reg_state = ppop();

  op = allocate(sizeof(op_t));
  op->opcode = NOP;
  op->value = (uval)reg_state;  /* save the register state */
  ppush(op);
  VVappend_def();

  ppush(op);
}

C( P: btok -- )
void VVcompile_again(void)  /* unconditional backwards branch */
{
  op_t *op1, *op2;

  op1 = ppop();

  ppush((ref)op1->value); VVload_arg_regs();

  op2 = allocate(sizeof(op_t));
  op2->opcode = GOTOQ;
  op2->r1 = ZERO;
  op2->label = op1;
  ppush(op2);
  VVappend_def();

  ppush((ref)op1->value); VVunlock_arg_regs();
}

C( P: btok -- )
void VVcompile_until(void)   /* conditional backwards branch */
{
  op_t *op1, *op2;
  reg_t reg;
  varbl *var;

  op1 = ppop();  
  ppush((ref)op1->value); VVload_arg_regs();  /* load the registers */
  ppush((ref)op1->value); VVunlock_random_regs();  /* get some registers to work with */

  ppush(&dstack); VVget_stack(); VVstack_pop(); var = ppop();
  ppush(var); VVvar_reg(); 
  reg = dpop();
  
  VVreg_state(); ppop();   /* get the register state */
                                       /* to get the registers into a consistent */
				       /* state */
  op2 = allocate(sizeof(op_t));
  op2->opcode = GOTOQ;
  op2->label = op1;
  op2->r1 = reg;
  ppush(op2);
  VVappend_def();
  
  ppush(var); VVrelease_var();
  ppush((ref)op1->value); VVunlock_normal_regs();
}


