#include "throf.h"
#include "dictionary.h"
#include "function.h"
#include "compile.h"
#include "memory.h"
#include "interpret.h"
#include "stack.h"
#include "colon-compiler.h"
#include "colon-stack.h"
#include "error.h"
#include "colon-branch.h"
#include "colon-state.h"
#include "colon-constants.h"
#include "colon-globals.h"

#if 0
static word_t * current_definition;
static op_t * last;
#endif

static op_t postlog = { RET, 0, 0, 0, 0, 0 ,0 ,0 ,0 };

void VVprep_for_compilation(void)
{
  varbl *var;
  VVnest_state();
  /* make sure the return stack pointer is in ESP */
  ppush(&rstack); VVget_stack(); var = ppop();
  ppush(var); dpush(ESP); VVvar_in_reg();
  /* maybe I should add a prologue here? */
}

#define CV_def  { \
  /* eax */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* ecx */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* edx */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* ebx */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* esp */ { GLOBAL_REG , {{ &rstack, offsetof(stack_data,place_pushed) }}}, \
  /* ebp */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* esi */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* edi */ { RANDOM_REG , {{ 0, 0 }}} \
	  } 

static arg_array CV = {CV_def};	    


void VVdo_compilation(void)
{
  func_t *func;
  op_t *end;
  colon_state *state;

  /* add the postlog */
  ppush(&CV); VVload_arg_regs();   /* flush all of the registers */

  end = allocate(sizeof(op_t));
  *end = postlog;
  ppush(end);
  VVappend_def();

  ppush(&CV); VVunlock_arg_regs();

  /* here is where I should call any/all code optimizers */
  /* well since I'm real lazy, I won't for now */

  VVget_state(); state = ppop();
  func = allocate(sizeof(func_t));
  func->func = compile_word(&state->current_definition);
  func->arg = CV;
  func->ret = CV;

  VVunnest_state();
  
  ppush(func);
}

/* make the current defintion call itself */
/* this is only momentarily correct,
   as soon as things stay in registers it needs to change */
void VVrecurse(void)
{
  op_t *op;
  colon_state *big_state;

  ppush(&CV); VVload_arg_regs();

  op = allocate(sizeof(op_t));
  op->opcode = GOTOQ;
  op->label = 0;
  op->r1 = ZERO;
  ppush(op);
  VVappend_def();

  ppush(&CV); VVunlock_arg_regs();
  
  VVget_state(); big_state = ppop();
  op->label = big_state->current_definition.first_op;
}

C( D: u -- )
void VVcompile_data_constant(void)
{
  uval value;
  varbl *stack, *constant, *new_const;

  value = dpop();
  
  dpush(value); VVget_constant(); constant = ppop();
  ppush(&dstack); VVget_stack(); stack = ppop();
  ppush(constant); ppush(stack); VVstack_push(); new_const = ppop();
}

C( P: p -- )
void VVcompile_pointer_constant(void)
{
  ref value;
  varbl *stack, *constant, *new_const;

  value = ppop();
  
  dpush((uval)value); VVget_constant(); constant = ppop();
  ppush(&pstack); VVget_stack(); stack = ppop();
  ppush(constant); ppush(stack); VVstack_push(); new_const = ppop();
}

static void load_arg_reg(arg_t *arg, reg_t reg)
{
  switch(arg->type)
    {
    case STACK_REG:
      {
	struct stack_reg *sr = &arg->arg.sr;
	/*
	   compile_pop of the appropiate stack (sr->stack) into the
	   appropriate register (sr->reg)  keep sr->reg locked!
	   */
	ppush(sr->stack);  VVget_stack();
	dpush(sr->depth); VVgrab_stack();
	dpush(reg); VVassign_var();
	break;
      }
    case GLOBAL_REG:
      {
	struct global_reg *gr = &arg->arg.gr;
	/*
	   compile load of the appropriate variable (gr->variable +
	   offset) into the appropriate register (gr->reg)
	   */	 
	ppush(gr->variable); dpush(gr->offset); VVget_global();
	dpush(reg); VVassign_var();
	break;
      }
    case RANDOM_REG:
      {  /* I must unload the register */
	/* so I load a dummy */
	dpush(reg); VVget_dummy(); dpush(reg); VVassign_var();
	break;
      }
    case PRESERVED_REG:
      {  /* in unload I'm not quite sure how to take advanage of this */
	/* so I'm not counting on it, pretending it's random */
	dpush(reg); VVget_dummy(); dpush(reg); VVassign_var();
	break;
      }
      break;  /* for now these are no ops */
    default:
      {
	char string[] = "sorry unknown argument passing type, aborting!!\n";
	dpush(sizeof(string) -1);
	ppush(string);
	VVfatal_error();
      }
    }
}

C( P: &arg_array -- )
void VVload_arg_regs(void)
{
  arg_array *array;
  arg_t *arg;
  uval i;
  uval flags[MAX_REG];
  array = ppop(); arg = &array->a[0];

  VVflush_stacks();  /* make pending actions happen */
  for(i = 0; i < MAX_REG; i++)  /* first set the non stack pointers */
    {
      uval flag;
      dpush(i); VVvar_for_reg(); VVis_stack_pointer(); flag = dpop();
      if (!flag)
	load_arg_reg(&arg[i], i);
      flags[i] = flag;
    }
  for(i = 0; i < MAX_REG; i++)  /* then set the stack pointers */
    {
      uval flag;
      dpush(i); VVvar_for_reg(); VVis_stack_pointer(); flag = dpop();
      if (flags[i])
	load_arg_reg(&arg[i], i);
    }
	
}

static void unlock_arg_reg(arg_t *arg, reg_t reg)
{
  switch(arg->type)
    {
    case STACK_REG:
      {
	struct stack_reg *sr = &arg->arg.sr;
	ppush(sr->stack);  VVget_stack();  /* get the right variable */
	dpush(sr->depth);  VVgrab_stack(); 
	VVrelease_var();   /* unlock the register assignment */
	break;
      }
    case GLOBAL_REG:
      {
	struct global_reg *gr = &arg->arg.gr;
	ppush(gr->variable); dpush(gr->offset); VVget_global();
	VVrelease_var();
	break;
      }
    case RANDOM_REG:
      {
	dpush(reg); VVget_dummy(); VVrelease_var();
	break;
      }
    case PRESERVED_REG:
      {
	/* again I can't count on taking advantage of this yet */
	dpush(reg); VVget_dummy(); VVrelease_var();
	break;
      }
    default:
      break;  /* for now these are no ops */
    }
}

C( P: &arg_array -- )
void VVunlock_random_regs(void)
{
  arg_array *array;
  uval i;
  array = ppop();
  for(i = 0; i < MAX_REG; i++)
    {
      if (array->a[i].type == RANDOM_REG)
	unlock_arg_reg(&array->a[i],i);
    }
}

C( P: &arg_array -- )
void VVunlock_normal_regs(void)
{
  arg_array *array;
  uval i;
  array = ppop();
  for(i = 0; i < MAX_REG; i++)
    {
      if (array->a[i].type != RANDOM_REG)
	unlock_arg_reg(&array->a[i],i);
    }
}

C( P: &arg_array -- )
void VVunlock_arg_regs(void)
{
  arg_array *array;
  uval i;
  array = ppop();
/* free the allocated registers */
  for(i = 0; i < MAX_REG; i++)
    unlock_arg_reg(&array->a[i],i);
}


C( P: &arg_array -- )
void VVunload_ret_regs(void)
{
  arg_array *array;
  arg_t *ret;
  uval i;
  array = ppop(); ret = &array->a[0];

  VVinvalidate_all();

/* get the return results where they go */
  for(i = 0; i < MAX_REG; i++)
    switch(ret[i].type)
      {
      case STACK_REG:
	{
	  struct stack_reg *sr = &ret[i].arg.sr;
	  /*
	  compile_push of the appropiate register (sr->reg) into the appropriate stack (sr->stack)
	  */
	  ppush(sr->stack);  VVget_stack();  /* get the right variable */
	  dpush(sr->depth);  VVgrab_stack(); 
	  dpush(i); VVvar_in_reg();
	  break;
	}
      case GLOBAL_REG:
	{
	  struct global_reg *gr = &ret[i].arg.gr;
	  /*
	     compile store of the appropriate register (gr->reg) into the appropriate 
	     variable (gr->variable + gr->offset)
	  */
	  ppush(gr->variable); dpush(gr->offset); VVget_global();
	  dpush(i); VVvar_in_reg();
	  break;
	}
      case RANDOM_REG:
      case PRESERVED_REG:
	break;  /* for now these are no ops */
      default:
	{
	  char string[] = "sorry unknown argument returning type, aborting!!\n";
	  dpush(sizeof(string) -1);
	  ppush(string);
	  VVfatal_error();
	}
      }
}

C( P: &func_t -- )
void VVcompile_call(void)  /* really stupid version */
{                          /* safe & easy not efficienty */
  func_t * func;
  op_t * op;

  func = ppop();

  ppush(&func->arg); VVload_arg_regs();

  op = allocate(sizeof(op_t));
  op->opcode = CALL;
  op->word = func;
  ppush(op);
  VVappend_def();

  ppush(&func->arg); VVunlock_arg_regs();
  ppush(&func->ret); VVunload_ret_regs();
}

void VVcolon_noname(void)
{
  state = COMPILING;
  VVprep_for_compilation();
}


void VVsemicolon(void)
{
  state = INTERPRETING;
  VVdo_compilation();
}


/* action of these words when not compiling is undefined */
C( D: u -- )   /* takes a word from the data stack and compiles 
                  instructions to push it back during execution */
void VVdconst(void)
{
  VVcompile_data_constant();
}

C( P: p -- )  /* takes a word from the pointer stack and compiles
                 instructions to push it back during execution */
void VVpconst(void)
{
  VVcompile_pointer_constant();
}

/*------------------ register the functions ---------------- */
void register_colon(void)
{
#define S(X) X , (sizeof(X) -1)

#define RET0 &VV
#define CALLs0 &VV

  register_function(S(":noname"),   FUNC_ADDR(VVcolon_noname),   0,   CALLs0, RET0);
  register_function(S(";"),         FUNC_ADDR(VVsemicolon),      IMM, CALLs0, RET0);
  register_function(S("dconst"),    FUNC_ADDR(VVdconst),         IMM, CALLs0, RET0);
  register_function(S("pconst"),    FUNC_ADDR(VVpconst),         IMM, CALLs0, RET0);
  register_function(S("recurse"),   FUNC_ADDR(VVrecurse),        IMM, CALLs0, RET0);
          /* from colon-branch */
  register_function(S("if"),        FUNC_ADDR(VVcompile_if),     IMM, CALLs0, RET0);
  register_function(S("ahead"),     FUNC_ADDR(VVcompile_ahead),  IMM, CALLs0, RET0);
  register_function(S("then"),      FUNC_ADDR(VVcompile_then),   IMM, CALLs0, RET0);
  register_function(S("begin"),     FUNC_ADDR(VVcompile_begin),  IMM, CALLs0, RET0);
  register_function(S("again"),     FUNC_ADDR(VVcompile_again),  IMM, CALLs0, RET0);
  register_function(S("until"),     FUNC_ADDR(VVcompile_until),  IMM, CALLs0, RET0);

#undef S
#undef CALLs0
#undef RET0
}

