#include "throf.h"
#include "spaghetti-stack.h"
#include "stack.h"
#include "function.h"
#include "compile.h"
#include "memory.h"
#include "colon-stack.h"
#include "colon-state.h"

#if !DISABLE_STACK_CONTINUATIONS
error
#endif

static void load_stack_val(void);

static void do_pending_pops(stack_state *state)
{
  stack_var *var;
  var = state->first;
  while(var != NIL)
    {
      if ((var->initial == NIL) && (var->popped_not_loaded))
	{
	  uval stack_locked = FALSE;
	  reg_t reg;
	  if ((state->this->reg == ZERO) || (state->this->flags & safe_to_remove ))
	    {
	      stack_locked = TRUE;
	      ppush(state->this); VVvar_reg(); reg=dpop();  /* load the register */
	      /* reg should now equal state->this->reg */
	    }
	  /* this isn't a register so I can't call load_stack_val */
	  /* it's not in a register because then it would be a form of pending */
	  /* push and I don't allow those here */
	  ppush(var->this); VVvar_reg(); reg=dpop();  
	  ppush(var->this); VVrelease_var();
	  if (stack_locked)
	    {
	      ppush(state->this); VVrelease_var(); /* say that later routines can flush it */
	    }
	}
      var = var->next;
    }
}

static void do_pending_pushes(stack_state *state)
{
  stack_var *var;
  var = state->first;
  while(var != NIL)
    {
      if (var->initial != NIL)
	{
	  uval stack_locked = FALSE;
	  if ((state->this->reg == ZERO) || (state->this->flags & safe_to_remove))
	    {
	      reg_t reg;
	      stack_locked = TRUE;
	      ppush(state->this); VVvar_reg(); reg=dpop();  /* load the register */
	      /* reg should now equal state->this->reg */
	    }
	  ppush(var->this); load_stack_val();

	  if (stack_locked)
	    {
	      stack_locked = FALSE;
	      ppush(state->this); VVrelease_var();
	    }
	}
      var = var->next;
    }
}

/* thinking, pushes cannnot be done until the stack pointer is adjusted up, but */
/* pops must be done before the stack pointer is adjusted down.  And all must */
/* have the stack pointer loaded before they can execute correctly */

static void do_adjustment(stack_state *state)
{
  do_pending_pops(state);  
  if (state->adjustment != 0)
    {
      op_t *op;
      stack_var *var;
      reg_t reg;
      
      var = state->first;
      while(var != NIL)
	{
	  var->depth += state->adjustment;
	  var = var->next;
	}
      if (state->this->reg == ZERO)
	{
	  ppush(state->this); VVvar_reg(); reg=dpop();  /* load the register */
	  ppush(state->this); VVrelease_var(); /* say that later routines can flush it */
	  /* reg should now equal state->this->reg */
	}
      
      op = allocate(sizeof(op_t));  /* verify this */
      if (state->adjustment > 0) /* push */
	{
	  op->opcode = (state->stack->direction == up)? ADD:SUB;
	  op->value = (state->adjustment)*sizeof(uval);
	}
      else /* pop */
	{
	  op->opcode = (state->stack->direction == up)? SUB:ADD;
	  op->value =  (-state->adjustment)*sizeof(uval);
	}
      op->r1 = state->this->reg;
      op->r2 = ZERO;
      ppush(op);
      VVappend_def();
      
      state->adjustment = 0;
    }
}

static void make_consistent_state(stack_state *state)
{
  do_adjustment(state);
  do_pending_pushes(state);

#if RETURN_STACK_STUCK
  if (state->stack == &rstack)
    return;
#endif
  ppush(state->this); VVflush_var();  /* make sure it's updated */
}

static void save_stack(void)
{
  stack_data *stack;
  stack_state *state;
  op_t *op;
  varbl *var;

  var = ppop(); state = var->v; stack = state->stack;

  do_adjustment(state);  /* make sure I have the correct value to save */

  op = allocate(sizeof(op_t));
  op->opcode = STORE;
  op->value = (uval)(&stack->place_pushed);
  op->r1 = ZERO;
  op->r2 = var->reg;
  ppush(op);
  VVappend_def();

#if !DISABLE_STACK_CONTINUATIONS
  op = allocate(sizeof(op_t));
  op->opcode = STORE;
  op->value = (uval)(&stack->place_to_pop);
  op->r1 = ZERO;
  op->r2 = var->reg;
  ppush(op);
  VVappend_def();
#endif
}

static void load_stack(void)
{
  stack_data *stack;
  stack_state *state;
  op_t *op;
  varbl *var;

  var = ppop(); state = var->v; stack = state->stack;
  
  op = allocate(sizeof(op_t));
  op->opcode = LOAD;
  op->value = (uval)(&stack->place_pushed);
  op->r1 = var->reg;
  op->r2 = ZERO;
  ppush(op);
  VVappend_def();
}

C( P: &arg_t &var -- )
static void preserve_stack(void)
{
  arg_t *arg;
  varbl *var;
  stack_state *state;

  var = ppop(); arg = ppop(); state = var->v;
  make_consistent_state(state);  /* make the state consistent for preservation */

  arg->type = GLOBAL_REG;
  arg->arg.gr.variable = state->stack;
  arg->arg.gr.offset = offsetof(stack_data,place_pushed);
}

static stack_state *get_stack_state(stack_data *stack)
{
  colon_state *big_state;
  stack_state *state;

  VVget_state(); big_state = ppop(); state = big_state->first_stack;

  /* attempt to find a preexisting state */
  while(state != NIL)
    {
      if (state->stack == stack)
	return state;
      state = state->next;
    }
  /* obviously I didn't so dynamically allocate one etc. */
  state = allocate(sizeof(stack_state));
  state->stack = stack;
  state->adjustment = 0;
  state->first = NIL;
  state->next = big_state->first_stack;
  big_state->first_stack = state;

  return state;
}

C( P: &stack_data -- &varbl )
void VVget_stack(void)
{
  stack_data *stack;
  stack_state *state;
  varbl *var;

  stack = ppop(); 
  state = get_stack_state(stack);

  ppush(&load_stack);  ppush(&save_stack);  ppush(&preserve_stack);
  dpush(0);  ppush(state);  VVnew_var();
  var = ppop(); state->this = var; ppush(var);
}

C( P: &varbl -- ) C( -- flag )
void VVis_stack_pointer(void)
{
  varbl *var;
  uval flag;
  var = ppop();
  flag = (var->load == &load_stack) &&     
         (var->save == &save_stack) &&
         (var->preserve == &preserve_stack);
  dpush(flag);
}

static void load_stack_val(void);
static void save_stack_val(void);
static void preserve_stack_val(void);

C( P: &stack_varbl -- &varbl ) C( depth -- )
void VVgrab_stack(void)
{
  varbl *vary, *vary2;
  stack_state *state;
  stack_var * var;
  sval depth;

  depth = dpop(); vary = ppop();  state = vary->v; var = state->first;
  while(var != NIL)
    {
      if (var->depth == (depth - state->adjustment))
	break;
      var = var->next;
    }
  if (var == NIL)
    {
      var = allocate(sizeof(stack_var));
      var->stack = state;
      var->next = state->first;
      var->depth = depth - state->adjustment;
      var->initial = NIL;
      var->popped_not_loaded = FALSE;
      state->first = var;

      ppush(&load_stack_val); 
      ppush(&save_stack_val); 
      ppush(&preserve_stack_val);
      dpush(0); ppush(var); VVnew_var();
      vary2 = ppop(); var->this = vary2;
    }
  ppush(var->this);
}

C( P: &stack_varbl -- &varbl ) 
void VVstack_pop(void)
{
  varbl *vary, *new_var;
  stack_state *state;
  stack_var *var;

  vary = ppop();  state = vary->v;
  ppush(vary); dpush(0); VVgrab_stack(); new_var = ppop(); var = new_var->v;

  /* change the stack adjustments */
  state->adjustment -= 1;
  
  /* tell it I'm a pop that hasn't */
  if (var->this->reg == ZERO)         /* if it's in a register it has */
    var->popped_not_loaded = TRUE;
  
  ppush(new_var);
}

static int is_stack_var(varbl *var)
{
  return ((var->load == &load_stack_val) &&
	  (var->save == &save_stack_val) &&
	  (var->preserve == &preserve_stack_val))?
	    TRUE : FALSE;
}

C( P: &varbl_initial_value &stack_varbl -- &varbl )
void VVstack_push(void)
{
  varbl *initial, *stackv, *new_var;
  stack_state *state;
  stack_var *var;

  stackv = ppop(); initial = ppop(); state = stackv->v;
  ppush(stackv); dpush(-1); VVgrab_stack(); new_var = ppop();

  /* change the stack adjustments */
  state->adjustment += 1;

  var = new_var->v;
  if (is_stack_var(initial))  /* code to prevent initial value loops */
    {                        /* it may break if other types are added */
      stack_var *var_init = initial->v;
      if (var_init->initial != NIL)
	{
	  if (var_init->initial != new_var)  /* make sure it's not myself */
	    initial = var_init->initial;
	  else
	    initial = NIL;
	}
    }
  var->initial = initial;
  ppush(new_var); VVinvalidate_var();  /* be sure it executes the load */
  ppush(new_var);
}

/* both push and pop has initial requirements that different from
   the usual stack variables.  Take special care.
*/


C( P: &varbl -- )
static
void load_stack_val(void)
{
  stack_var * var;
  varbl *vary;
  op_t *op;

  vary = ppop(); var = vary->v;
  
  if (var->initial != NIL)  /* this is from a push, I need to take special action */
    {
      ppush(var->initial); ppush(vary); VVtransfer_var();
      var->initial = NIL;         /* don't do this again */
      var->popped_not_loaded = FALSE;  /* just incase it's a case of push pop */
    }
  else if ((var->popped_not_loaded) || (var->depth + var->stack->adjustment))
    {
      uval stack_locked = FALSE;
      var->popped_not_loaded = FALSE;

      if (var->depth < 0)  /* if I haven't moved the stack pointer so I'm on the stack */
	do_adjustment(var->stack);   /* it might be better to signal an error */
                                     /* I can't think of a case where this could */
				     /* happen.  The value is likely trashed */ 

      if (var->stack->this->reg == ZERO) /* this is unlikely, oh well handle it */
	{
	  stack_locked = TRUE;
	  ppush(var->stack->this); VVvar_reg(); dpop(); /* load it in  a register */
	  /* the value just dropped should equal var->stack->this->reg */
	}

      op = allocate(sizeof(op_t));
      op->opcode = LOAD;
      op->r1 = var->this->reg;
      op->r2 = var->stack->this->reg;
      if (var->stack->stack->direction == down)
	op->value = var->depth*sizeof(uval);
      else
	op->value = - (var->depth*sizeof(uval));  /* ah, twos complement addition :) */
      ppush(op);  VVappend_def();

      if (stack_locked) 
	{
	  ppush(var->stack->this); VVrelease_var(); /* allow later routines to */
	                                            /* flush it */
	}
    }
}


C( P: &varbl -- )
static
void save_stack_val(void)
{
  stack_var * var;
  varbl *vary;
  op_t *op;

  vary = ppop(); var = vary->v;
  
  if ((var->depth + var->stack->adjustment) >= 0)  /* if the stack var is really on the stack */
    {
      uval stack_locked = FALSE;
      if (var->depth < 0)  /* if I haven't moved the stack pointer so I'm on the stack */
	do_adjustment(var->stack);

      if (var->stack->this->reg == ZERO) /* this is unlikely, oh well handle it */
	{
	  stack_locked = TRUE;
	  ppush(var->stack->this); VVvar_reg(); dpop(); /* load the register */
	  /* the value just dropped should equal var->stack->this->reg */
	}
      op = allocate(sizeof(op_t));
      op->opcode = STORE;
      op->r1 = var->stack->this->reg;
      op->r2 = var->this->reg;
      if (var->stack->stack->direction == down)
	op->value = var->depth*sizeof(uval);
      else
	op->value = -(var->depth*sizeof(uval));  /* ah, twos complement addition :) */
	  
      ppush(op);
      VVappend_def();
      if (stack_locked)
	{
	  ppush(var->stack->this); VVrelease_var();     /* allow later routines */
	                                               /* to flush it */
	}
    }
}


C( P: &arg_t &var -- )
static void preserve_stack_val(void)
{
  arg_t *arg;
  varbl *vary;
  stack_var *var;

  vary = ppop(); arg = ppop(); var = vary->v;
  make_consistent_state(var->stack);  /* make it all ready for preservation */

  if (var->depth < 0)  /* if I'm not on the stack I can't be preserved */
    {
      arg->type = RANDOM_REG;
      arg->arg.rr.data1 = 0;
      arg->arg.rr.data2 = 0;
    }
  else 
    {
      /* stack depth change is to hard to compute */
      /* simply insist on flushing, and forgetting about it */
      ppush(vary); save_stack_val();  /* save my value */
      vary->flags &= ~safe_to_remove;  /* lock myself in, to preserve */
				       /* consistency */

      arg->type = RANDOM_REG;  /* and say I'm a random register */
      arg->arg.rr.data1 = 0;
      arg->arg.rr.data2 = 0;
#if 0
      arg->type = STACK_REG;
      arg->arg.sr.stack = var->stack->stack;
       /* var->stack->adjustment == 0 */
      arg->arg.sr.depth = var->depth + var->stack->adjustment;
#endif
    }
}

void VVflush_stacks(void)
{
  stack_state *state;
  varbl *var;
  /* I really should keep track of them instead of doing it statically */
  ppush(&dstack); VVget_stack(); var = ppop(); state = var->v; 
  make_consistent_state(state);
  ppush(&pstack); VVget_stack(); var = ppop(); state = var->v; 
  make_consistent_state(state);
  ppush(&fstack); VVget_stack(); var = ppop(); state = var->v; 
  make_consistent_state(state);
  ppush(&rstack); VVget_stack(); var = ppop(); state = var->v; 
  make_consistent_state(state);
}

#if 0  /* old code */

/* in these two I implicitly assume that,
   the operation is safe and doesn't need checking.
    This can't last,  more work !!! 
*/

C( P: &varbl_to &varbl_stack -- )
void VVcompile_pop(void)
{
  op_t * op;
  reg_t stack;
  reg_t value;
  stack_data * stack_dat;
  varbl * stack_v;

  stack_v = ppop(); stack_dat = stack_v->v; 
  ppush(stack_v); VVvar_reg(); stack = dpop();
  VVvar_reg(); value = dpop();
  
  op = allocate(sizeof(op_t));
  op->opcode = LOAD;
  op->r1 = value;
  op->r2 = stack;
  op->value = 0;
  ppush(op);
  VVappend_def();

  op = allocate(sizeof(op_t));
  op->opcode = (stack_dat->direction == up)? SUB : ADD;
  op->r1 = stack;
  op->r2 = ZERO;
  op->value = sizeof(uval);
  ppush(op);
  VVappend_def();
}

C( P: &varbl_to &varbl_stack -- )
void VVcompile_push(void) 
{
  op_t * op;
  reg_t stack;
  reg_t value;
  stack_data *stack_dat;
  varbl * stack_v;

  stack_v = ppop(); stack_dat = stack_v->v; 
  ppush(stack_v); VVvar_reg(); stack = dpop();
  VVvar_reg(); value = dpop();
  
  op = allocate(sizeof(op_t));
  op->opcode = (stack_dat->direction == up)? ADD: SUB;
  op->r1 = stack;
  op->r2 = ZERO;
  op->value = sizeof(uval);
  ppush(op);
  VVappend_def();

  op = allocate(sizeof(op_t));
  op->opcode = STORE;
  op->r1 = stack;
  op->r2 = value;
  op->value = 0;
  ppush(op);
  VVappend_def();
}

#endif /* old code */
