#include "colon-variable.h"
#include "compile.h"
#include "stack.h"
#include "memory.h"
#include "throf.h"
#include "colon-state.h"
#include "colon-constants.h"

static struct colon_state *cur_state = NIL;

C( P: -- &colon_state )
void VVget_state(void)
{
  ppush(cur_state);
}


void VVnest_state(void)
{
  colon_state *new_state;
  
  new_state = allocate(sizeof(colon_state));
  new_state->current_definition.first_op = 0;
  new_state->last = 0;
  init_state(&new_state->rstate);
  new_state->prev = cur_state;
  new_state->first_stack = NIL;

  cur_state = new_state;
}

void VVunnest_state(void)
{
  if (cur_state != NIL)
    {
      cur_state = cur_state->prev;
    }
}

C( P: &op_t -- ) 
void VVappend_def(void)  /* this handles all of the pointer stuff */
{
  op_t *next = ppop();
  if (cur_state->last == 0) {
    cur_state->current_definition.first_op = next;
    next->prev = 0;
  } else {
    cur_state->last->next = next;
    next->prev = cur_state->last;
  }
  cur_state->last = next;
  next->next = 0;
}

C( u -- ) C( P: &load &save v  -- &varbl)
void VVnew_var(void)
{
  varbl *var;
  variable_method load,save,preserve;
  uval u;
  void *v;
  v = ppop();  u = dpop(); preserve = ppop(); save = ppop(); load = ppop();
  var = new_variable(&cur_state->rstate,load,save,preserve,u,v);
  ppush(var);
}

C( P: &var -- ) C( -- reg )
void VVvar_reg(void)
{
  varbl *var;
  reg_t reg;
  var = ppop();
  reg = reg_for_variable(&cur_state->rstate, var);
  dpush(reg);
}

C( P: &var -- ) C( reg -- )
void VVassign_var(void)
{
  reg_t reg;
  varbl *var;
  reg = dpop();
  var = ppop();
  assign_variable(&cur_state->rstate, var, reg);
}

C( P: &var -- )
void VVrelease_var(void)
{
  varbl *var;
  var = ppop();
  release_variable(&cur_state->rstate, var);
}

C( P: &var_from &var_to -- )
void VVtransfer_var(void)
{
  varbl *from, *to;
  to = ppop(); from = ppop();
  transfer_value(&cur_state->rstate, from, to);
}

C( P: &var -- )
void VVinvalidate_var(void)
{
  varbl *var;
  var = ppop();
  invalidate_variable(&cur_state->rstate, var);
}

C( -- )
void VVinvalidate_all(void)
{
  invalidate_all(&cur_state->rstate);
}

C( P: &var -- ) C( reg -- )
void VVvar_in_reg(void)
{
  varbl *var;
  reg_t reg;
  var = ppop(); reg = dpop();
  in_variable(&cur_state->rstate,var,reg);
}

C( P: -- &arg_array )
void VVreg_state()
{
  ppush(get_register_state(&cur_state->rstate));
}

C( reg -- ) C( P: -- &varbl )
void VVvar_for_reg(void)
{
  reg_t reg;
  varbl *var;
  reg = dpop();
  var = cur_state->rstate.regs[reg];
  if (var == NIL)
    {                /* put a dummy in to fill the gap */
      dpush(reg); VVget_dummy(); var = ppop();
      ppush(var); dpush(reg); VVassign_var(); 
      ppush(var); VVrelease_var(); 
    }
  ppush(var);
}

C( P: &varbl -- )
void VVflush_var(void)
{
  varbl *var;
  var = ppop();
  flush_variable(&cur_state->rstate, var);
}
