#ifndef COMPILE_H
#define COMPILE_H

#include "throf.h"
#include "function.h"

typedef enum opcodes { 
  RET,     /* syntax: ret                 || returns to caller                       */
  CALL,    /* syntax: call word           || transfers control to word               */
  JMP,     /* syntax: jmp  word           || transfers control to word               */
  GOTOQ,   /* syntax: gotoq r1,label      || transfers control to label, if reg == 0 */
  LOAD,    /* syntax: load  r1,[r2+value] || r1 <- [r2+value]                        */
  STORE,   /* syntax: store [r1+value],r2 || [r1+value] <- r2                        */
  SET,     /* syntax: set   r1,r2+value   || r1 <- r2 + value                        */
  XCHG,    /* syntax: xchg  r1,r2         || r1 <- r2, r2 <- r1                      */
  ADD,     /* syntax: add   r1,r2+value   || r1 <- r1 + (r2 + value)                 */
  SUB,     /* syntax: sub   r1,r2+value   || r1 <- r1 - (r2 + value)                 */
  SYSCALL, /* syntax: syscall value       || execute system call value               */
  NOP,     /* syntax: nop                 || does nothing put a label on it          */
  LAST_OPCODE } opcodes;

typedef struct word word_t; /* word_t defined later */
typedef char * label_t;     /* tenative definition */
typedef void * var_t;       /* tenative definition */

/* I should probably change the operation structure to a contigous allocated thing */
typedef struct operation {
  opcodes            opcode;
  uval               offset;
  reg_t              r1;
  reg_t              r2;
  uval               value;
  struct operation * label;
  func_t *           word;
  struct operation * next;
  struct operation * prev;
} op_t ;

struct word {
  op_t *first_op;  /* the used part */
};

func_ptr_t compile_word(word_t *word);

#endif COMPILE_H
