#include "throf.h"
#include "stack.h"
#include "dictionary.h"
#include "colon-compiler.h"
#include "colon-branch.h"
#include "colon-stack.h"
#include "core-compiler.h"
#include "interpret.h"  /* I probably shouldn't */
/* define all of the primitives I need for a compiler */

/* double number management routines */
uval low(udval x)
{
  return x & UVAL_MAX;
}

uval high(udval x)
{
  return x >> UVAL_BITS;
}

udval join(uval low,uval high)
{
  udval x;
  x = high;
  x = x << UVAL_BITS | low;
  return x;
}

void ddpush(udval x)
{
  dpush(high(x));
  dpush(low(x));
}

udval ddpop()
{
  uval high;
  uval low;
  high = dpop();
  low = dpop();
  return join(low,high);
}

/* memory access words */

void bang(void)
{
  uval val;
  uval * p;
  p = (uval *)dpop();
  val = dpop();
  *p = val;
}

void at(void)
{
  uval *p;
  p = (uval *) dpop();
  dpush(*p);
}

/* math words */

void plus(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush(u1+u2);
}

void minus(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush(u1 -u2);
}

void UM_star(void)
{
  uval u1, u2;
  udval ud;
  
  u2 = dpop();
  u1 = dpop();
  
  ud = u1;
  ud *= u2;
  ddpush(ud);
}

void UM_slash_MOD(void)
{
  udval d; uval u1;
  uval rem, quot;
  
  u1 = dpop();
  d = ddpop();
  
  rem = d % u1;
  quot = d / u1;

  dpush(rem);
  dpush(quot);
}

void M_star(void)
{
  sval n1, n2;
  sdval nd;
  
  n2 = dpop();
  n1 = dpop();
  
  nd = n1;
  nd *= n2;
  ddpush(nd);
}

void FM_slash_MOD(void)
{
  sdval sd;  uval n1;
  uval rem, quot;
  
  sd = ddpop();
  n1 = dpop();

  { 
    int ssd, sn1;  /* take the signs & make everthing positive */
    if (sd < 0 ) {sd = -sd; ssd = -1;} else { ssd = 1;}
    if (n1 < 0 ) {n1 = -n1; sn1 = -1;} else { sn1 = 1;}

    if (ssd != sn1)
      {
	rem = n1 - sd % n1;
	quot = -((sd + n1 - 1) / n1);
      }
    else
      {
	rem = sd % n1;
	quot = sd / n1;
      }
  }
    dpush(rem);
    dpush(quot);
}


void SM_slash_REM(void)
{
  sdval sd;  uval n1;
  uval rem, quot;
  
  sd = ddpop();
  n1 = dpop();

  { 
    int ssd, sn1;  /* take the signs & make everthing positive */
    if (sd < 0 ) {sd = -sd; ssd = -1;} else { ssd = 1;}
    if (n1 < 0 ) {n1 = -n1; sn1 = -1;} else { sn1 = 1;}

    rem = sd % n1;
    quot = sd / n1;

    if (ssd != sn1)
      {
	rem = -rem;
	quot = -quot;
      }
  }

    dpush(rem);
    dpush(quot);
}


void equal(void)
{
  uval u1, u2;
  u2 = dpop();
  u1 = dpop();
  dpush((u1 == u2)?TRUE:FALSE);
}

void not_equal(void)
{
  uval u1, u2;
  u2 = dpop();
  u1 = dpop();
  dpush((u1 != u2)?TRUE:FALSE);
}

void less(void)
{
  sval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush((n1 < n2)?TRUE:FALSE);
}

void greater(void)
{
  sval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush((n1 > n2)?TRUE:FALSE);
}

void U_less(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush((u1 < u2)?TRUE:FALSE);
}

void U_greater(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush((u1 > u2)?TRUE:FALSE);
}

/* boolean logic */
void AND(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush(u1&u2);
}

void OR(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush(u1|u2);
}

void XOR(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush(u1^u2);
}

void LSHIFT(void)
{
  uval x, u;
  u = dpop();
  x = dpop();
  dpush( x << u);
}

void RSHIFT(void)
{
  uval x, u;
  u = dpop();
  x = dpop();
  dpush( x >> u);
}

void ARSHIFT(void)
{
  sval x; uval u;
  u = dpop();
  x = dpop();
  dpush( (x >=0)? ( x >> u) : ~((~x) >> u) );
}

/* my compiler specific ones */
void COMPILE_comma(void)  {  ppush((ref)dpop()); VVcompile_call(); }
void EXECUTE(void)  {  ppush((ref)dpop()); execute_function();  }
void LITERAL(void)  {  VVdconst();  }
void colon_NONAME(void)  {   VVcolon_noname(); }
void semicolon(void)  {    VVsemicolon();  dpush((uval)ppop()); }
void RECURSE(void)  {   VVrecurse(); }
void IF(void)  { VVcompile_if(); }
void AHEAD(void) { VVcompile_ahead(); }
void THEN(void) { VVcompile_then(); }
void BEGIN(void) { VVcompile_begin(); }
void AGAIN(void) { VVcompile_again(); }
void UNTIL(void) { VVcompile_until(); }

/* little utility defintions */
void CS_ROLL(void)
{
  uval place;
  place = dpop();
  proll(place);
}

void CS_PICK(void)
{
  uval place;
  place = dpop();
  ppush(ppeek(place));
}

/* stack words */

void ROLL(void)
{
  uval u = dpop();
  droll(u);
}

void PICK(void)
{
  uval u = dpop();
  dpush(dpeek(u));
}

void DROP(void)
{
  if (state == INTERPRETING)
   {  dpop(); }
  else
   {  ppush(&dstack); VVget_stack(); VVstack_pop(); ppop(); }
}

void DEPTH(void)
{
  dpush(ddepth());
}

void R_to(void)
{
  varbl *var;

  /* get a value off the return stack */
  ppush(&rstack); VVget_stack(); VVstack_pop(); var = ppop();
  /* now push it on the data stack */
  ppush(var); ppush(&dstack); VVget_stack(); VVstack_push(); var = ppop();
}

void to_R(void)
{
  varbl *var;

  /* get a value off the data stack */
  ppush(&dstack); VVget_stack(); VVstack_pop(); var = ppop();
  /* now push it on the return stack */
  ppush(var); ppush(&rstack); VVget_stack(); VVstack_push(); var = ppop();
}

#if 0
void to_R(void)
{
  reg_t val_reg;

  /* allocate a register to use */
  VVget_register(); val_reg = dpop();
  /* allocate the data stack pointer to a register */
  ppush(&dstack);  VVget_stack(); 

  /* pop the data stack */
  dpush(val_reg); ppush(&dstack); VVcompile_pop();

  /* push the value on the return stack */
  dpush(val_reg); ppush(&rstack); VVcompile_push();

  /* free the register */
  dpush(val_reg); VVfree_register();
  /* free the data stack pointer */
  ppush(&dstack); VVflush_stack_pointer();
}
#endif

/*----------------------  Start registration ------------------------------*/

void register_core_compiler(void)
{
#define S(X) X , (sizeof(X) -1)

#define RET0 &VV
#define CALLs0 &VV

  register_function(S("ANS!"),     FUNC_ADDR(bang),             0,   CALLs0, RET0);
  register_function(S("ANS@"),     FUNC_ADDR(at),               0,   CALLs0, RET0);
  register_function(S("M*"),       FUNC_ADDR(M_star),           0,   CALLs0, RET0);
  register_function(S("UM*"),      FUNC_ADDR(UM_star),          0,   CALLs0, RET0);
  register_function(S("UM/MOD"),   FUNC_ADDR(UM_slash_MOD),     0,   CALLs0, RET0);
  register_function(S("+"),        FUNC_ADDR(plus),             0,   CALLs0, RET0);
  register_function(S("-"),        FUNC_ADDR(minus),            0,   CALLs0, RET0);
  register_function(S("FM/MOD"),   FUNC_ADDR(FM_slash_MOD),     0,   CALLs0, RET0);
  register_function(S("SM/REM"),   FUNC_ADDR(SM_slash_REM),     0,   CALLs0, RET0);
  register_function(S("<"),        FUNC_ADDR(less),             0,   CALLs0, RET0);
  register_function(S("="),        FUNC_ADDR(equal),            0,   CALLs0, RET0);
  register_function(S(">"),        FUNC_ADDR(greater),          0,   CALLs0, RET0);
  register_function(S("<>"),       FUNC_ADDR(not_equal),        0,   CALLs0, RET0);
  register_function(S("U<"),       FUNC_ADDR(U_less),           0,   CALLs0, RET0);
  register_function(S("U>"),       FUNC_ADDR(U_greater),        0,   CALLs0, RET0);
  register_function(S("AND"),      FUNC_ADDR(AND),              0,   CALLs0, RET0);
  register_function(S("OR"),       FUNC_ADDR(OR),               0,   CALLs0, RET0);
  register_function(S("XOR"),      FUNC_ADDR(XOR),              0,   CALLs0, RET0);
  register_function(S("LSHIFT"),   FUNC_ADDR(LSHIFT),           0,   CALLs0, RET0);
  register_function(S("RSHIFT"),   FUNC_ADDR(RSHIFT),           0,   CALLs0, RET0);
  register_function(S("ARSHIFT"),  FUNC_ADDR(ARSHIFT),          0,   CALLs0, RET0);
  
  register_function(S("COMPILE,"), FUNC_ADDR(COMPILE_comma),    0,   CALLs0, RET0);
  register_function(S("EXECUTE"),  FUNC_ADDR(EXECUTE),          0,   CALLs0, RET0);
  register_function(S("LITERAL"),  FUNC_ADDR(LITERAL),          IMM, CALLs0, RET0);
  register_function(S(":NONAME"),  FUNC_ADDR(colon_NONAME),     0,   CALLs0, RET0);
  register_function(S("ANS;"),     FUNC_ADDR(semicolon),        IMM, CALLs0, RET0);
  register_function(S("RECURSE"),  FUNC_ADDR(RECURSE),          IMM, CALLs0, RET0);
  register_function(S("IF"),       FUNC_ADDR(IF),               IMM, CALLs0, RET0);
  register_function(S("AHEAD"),    FUNC_ADDR(AHEAD),            IMM, CALLs0, RET0);
  register_function(S("THEN"),     FUNC_ADDR(THEN),             IMM, CALLs0, RET0);
  register_function(S("BEGIN"),    FUNC_ADDR(BEGIN),            IMM, CALLs0, RET0);
  register_function(S("AGAIN"),    FUNC_ADDR(AGAIN),            IMM, CALLs0, RET0);
  register_function(S("UNTIL"),    FUNC_ADDR(UNTIL),            IMM, CALLs0, RET0);
  register_function(S("CS-PICK"),  FUNC_ADDR(CS_PICK),          0,   CALLs0, RET0);
  register_function(S("CS-ROLL"),  FUNC_ADDR(CS_ROLL),          0,   CALLs0, RET0);

  register_function(S("ROLL"),     FUNC_ADDR(ROLL),             0,   CALLs0, RET0);
  register_function(S("PICK"),     FUNC_ADDR(PICK),             0,   CALLs0, RET0);
  register_function(S("DEPTH"),    FUNC_ADDR(DEPTH),            0,   CALLs0, RET0);
  register_function(S("DROP"),     FUNC_ADDR(DROP),             IMM, CALLs0, RET0);

  register_function(S("R>"),       FUNC_ADDR(R_to),             IMM, CALLs0, RET0);
  register_function(S(">R"),       FUNC_ADDR(to_R),             IMM, CALLs0, RET0);

#undef S
#undef CALLs0
#undef RET0
}
