#include <stdio.h>
#include "throf.h"
#include "string.h"
#include "function.h"
#include "dictionary.h"
#include "memory.h"
#include "stack.h"

#define DEBUG
#undef DEBUG

void nop(void);
func_t nop_func = {FUNC_ADDR(nop),{VV_def},{VV_def}}; 
func_node nop_node = { { 3,"nop",0,&nop_func},0,0};

func_node * root = &nop_node;

void nop(void)
{ return; }


void traverse_inorder(void (*func)(func_node *,int), func_node *start, int number)
{
  if (start != 0)
    {
      traverse_inorder(func,start->less,2*number);
      func(start,number);
      traverse_inorder(func,start->greater,2*number+1);
    }
}

void traverse_preorder(void (*func)(func_node *,int), func_node *start, int number)
{
  if (start != 0)
    {
      func(start, number);
      traverse_preorder(func,start->less,2*number);
      traverse_preorder(func,start->greater,2*number+1);
    }
}

#ifdef DEBUG
unsigned node_count;
void display_node(func_node * node, int number)
{
  int i;
  node_count++;
  printf("node(%p): %d, less(%p), greater(%p) \n",node,number,node->less,node->greater);
  printf("name(%p) = ",node->entry.name);
  for(i = 0;i < node->entry.name_len; i++)
    putchar(node->entry.name[i]);
  printf("\nfunc = %p, flags = %d\n", node->entry.execution->func,node->entry.flags);
  printf("\n\n");
}
#endif

#ifdef DEBUG
void show_dictionary(void)
{
  node_count = 0;
  printf("Dictionary:\n");
  traverse_inorder(display_node,root,1);
  printf("node_count = %d\n---------------------------------------\n",node_count);
  node_count = 0;
  fflush(stdout);
}
#endif


func_node * find_function(char * str,unsigned str_len, func_node * start)
{
  int comp;

#ifdef DEBUG
  show_dictionary();
#endif

  while(start != 0)
    {
      comp = compare(str, str_len, start->entry.name, start->entry.name_len);
      if (comp == 0)
	break;
      else if (comp > 0)
	start = start->greater;
      else /* comp < 0 */
	start = start->less;
    }
  return start;
}

entry * find_func(char * str, unsigned str_len)
{
  func_node * here;
  here = find_function(str,str_len,root);
  return (here != 0)?&here->entry:0;
}

void add_node(func_node *new)
{
  int comp;
  func_node * start, *last, *real_last;

#ifdef DEBUG
  show_dictionary();
#endif

  start = root;

  new->less = 0;
  new->greater = 0;
  last = start;

  /* find where to insert the new node */
  while(start != 0)  /* it is guaranteed that this loop executes at least once! */
    {
      real_last = last;
      last = start;
      comp = compare(new->entry.name, new->entry.name_len, 
		     start->entry.name, start->entry.name_len);
      if (comp == 0)
	break;
      else if (comp > 0)
	start = start->greater;
      else /* comp < 0 */
	start = start->less;
    }
  if (comp > 0)
    last->greater = new;
  else if (comp < 0)
    last->less = new;
  else /* if comp == 0) */
    {
      new->greater = start;
      new->less = start->less;
      start->less = 0;
      if (real_last->greater == start)
	  real_last->greater = new;
      else /* if (real_last->less == start) */
	real_last->less = new;
  }

#ifdef DEBUG
  show_dictionary();
#endif

}

entry *register_func(char * name, unsigned name_len, func_t *func, func_flags_t flags)
{
  int i;
  
  func_node *node;

#ifdef DEBUG
  show_dictionary();
#endif

  node = allocate(sizeof(func_node));
  node->entry.name_len = name_len;
  if (name_len > 0)
    node->entry.name = allocate(sizeof(char)*name_len);
  else
    node->entry.name = 0;
  node->entry.execution = func;
  node->entry.flags = flags;

  for(i = 0; i < name_len; i++)
    {
      node->entry.name[i] = name[i];
    }

  add_node(node);

  return &node->entry;
}


void register_function(char * name, unsigned name_len, func_ptr_t func, 
		       func_flags_t flags,
		       arg_array *arg, arg_array *ret)
{
  func_t * func_p;

#ifdef DEBUG
  show_dictionary();
#endif

  func_p = allocate(sizeof(func_t));
  func_p->func = func;
  func_p->arg = *arg;
  func_p->ret = *ret;

  register_func(name,name_len,func_p,flags);
}

