#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#include "throf.h"
#include "function.h"
#include "spaghetti-stack.h"
#include "stack.h"

arg_array VV = { VV_def };

/* since this is really the home of what is in function.h */
struct  do_call_params {
  unsigned regs[MAX_REG]; /* 0 */  /* array of registers, ignore esp */
  /* register offset
         eax  0
	 ecx  4
	 edx  8
	 ebx  12
	 esp  16
	 ebp  20
	 esi  24
	 edi  28
   */
  void * function;      /* 32 */
  unsigned reserved;    /* 36 */
};

void do_call(struct do_call_params *params);

void execute_func(void);

#define RETURN_BIAS 1
/* return bias is the number of words pushed in between the start of */
/* execute_function and sync_return_stack */

void execute_function(void) 
{
  sync_return_stack();  /* so I can grab things off of it */
  execute_func();
}
void execute_func(void)
{
  func_t *func;

  /* first calculate the values to go in all of the registers */
  struct do_call_params params;
  unsigned *regs = &params.regs[0];
  unsigned i;

  func = ppop();
  for(i = 0; i < MAX_REG; i++)
      switch(func->arg.a[i].type)
      {
      case STACK_REG:
	{
	  struct stack_reg *sr = &func->arg.a[i].arg.sr;
	  if ((sr->stack) == &rstack) 
	    regs[i] = peek(sr->depth+RETURN_BIAS,&rstack);
	  else
	    regs[i] = peek(sr->depth,sr->stack);
	  break;
	}
      case GLOBAL_REG:
	{
	  struct global_reg *gr = &func->arg.a[i].arg.gr;
	  unsigned char * var;
	  var = gr->variable;
	  regs[i] = *((unsigned *)(var +gr->offset));
	  break;
	}
      case RANDOM_REG:
	{
	  struct random_reg *rr = &func->arg.a[i].arg.rr;
	  /* ignore this nothing to do :) */
	  regs[i] = 0;  /* just for cleanliness */
	  break;
	}
      case PRESERVED_REG:
	{
	  struct preserved_reg *pr = &func->arg.a[i].arg.pr;
	  /* ignore this nothing to do :) */
	  regs[i] = 0;   /* just for cleanliness */
	  break;
	}
      default:
	goto error;
      }
  params.function = func->func;
  do_call(&params);
#if DISABLE_STACK_CONTINUATIONS
  /* only place_pushed really counts */
  dstack.place_to_pop = dstack.place_pushed;
  rstack.place_to_pop = dstack.place_pushed;
  pstack.place_to_pop = pstack.place_pushed;
  fstack.place_to_pop = fstack.place_pushed;
#endif
  for(i = 0; i < MAX_REG; i++)
      switch(func->ret.a[i].type)
      {
      case STACK_REG:
	{
	  struct stack_reg *sr = &func->ret.a[i].arg.sr;
#if DISABLE_STACK_CONTINUATIONS
	  if ((sr->stack) == &rstack)
	    poke(sr->depth+RETURN_BIAS,regs[i],&rstack);
	  else
	    poke(sr->depth,regs[i],sr->stack);
#else
#error 
#endif
	  break;
	}
      case GLOBAL_REG:
	{
	  struct global_reg *gr = &func->arg.a[i].arg.gr;
	  unsigned char * var;
	  var = gr->variable;
	  *((unsigned *)(var +gr->offset)) = regs[i];
	  break;
	}
      case RANDOM_REG:
	{
	  struct random_reg *rr = &func->arg.a[i].arg.rr;
	  /* ignore this nothing to do :) */
	  break;
	}
      case PRESERVED_REG:
	{
	  struct preserved_reg *pr = &func->arg.a[i].arg.pr;
	  /* ignore this nothing to do :) */
	  break;
	}
      default:
	goto error;
      }

  return;
 error:
  {
    fprintf(stderr,"sorry unknown argument passing type, aborting!!\n");
    exit(EXIT_FAILURE);
  }

}


#if 0
void do_call(struct do_call_params params)
{
  struct do_call_params *params_copy;
  params_copy = alloca(sizeof(struct do_call_params));
  *params_copy = *params;
  params_copy->reserved = (unsigned)&&return_to; /* a nice gcc extention */
#define EOL " \n\t"
  asm(
      "pop %eax" EOL
      "pop %ecx" EOL
      "pop %edx" EOL
      "pop %ebx" EOL
      "add $4,%esp" EOL /* ignore %esp */
      "pop %ebp" EOL
      "pop %esi" EOL
      "pop %edi" EOL
      "ret" EOL);
 return_to:
  asm(
      "push 0" EOL    /* dummy to fill in reserved1 */
      "push 0" EOL    /* dummy to fill in function */
      "push %edi" EOL
      "push %esi" EOL
      "push %ebp" EOL
      "push %esp" EOL /* like this value will be useful :) */
      "push %ebx" EOL
      "push %edx" EOL
      "push %ecx" EOL
      "push %eax" EOL
      );
  *params = *params_copy;
}
#endif
