#ifndef FUNCTION_H
#define FUNCTION_H

#include "throf.h"
#include <stddef.h>
#include "stack.h"

typedef void (*func_ptr_t)(void);

#define FUNC_ADDR(X) ((func_ptr_t)(&(X)))

typedef enum arg_type { 
STACK_REG, GLOBAL_REG, RANDOM_REG, PRESERVED_REG
/* STACK_FLUSH, GLOBAL_FLUSH */
} arg_type;

typedef enum regs { EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI, ZERO } reg_t;
#define MAX_REG (ZERO)
/* ZERO isn't a necesarrily a real register, it makes a useful noop */

/* What the garbage collector will get to see :) */
/* Also what the structure is statically initialized with */
struct generic {
  ref  data1;
  uval data2;
/*  uval data3; */
};

/* Note a function which passes and returns a variable in a register   */
/* is not responsible for that register being updated in it's          */
/* `permanent' home.                                                   */
/* Additionally for now at least it is an error if a function takes    */
/* stack argmnent n, when n!= 0, if it has not take stack argment n-1. */

/* A stack variable passed in a register */
/* or A stack variable is returned in a register */
struct stack_reg {
  struct stack_data * stack;
  uval depth;
/*  reg_t reg;   */
};

/* A global variable passed in a register */
struct global_reg {
  ref variable;
  uval offset;
/*  reg_t reg; */
};

/* A random (trashed register) */
struct random_reg {
  ref  data1; /* must be 0 */
  uval data2; /* must be 0 */
/*  reg_t reg; */
};

/* A preseved register, across the call */
struct preserved_reg {
  ref  data1;  /* must be 0 */
  uval data2;  /* must be 0 */
/*  reg_t reg;   */
};

#if 0  
/* for statements of which globals should be sure to be flushed before */
/* a call goes through is silly, because it implies otherwise they */
/* don't have to be flushed.  This requires the complete and entire */
/* call chain to be known, which possibly is very silly.  Also I'm */
/* reducing this to simply register management an call time so this */
/* kind of information should be handled elsewhere */

/* A stack should be sure to be in it's `permanent' position when this */
/* call occurs.  Possibly I should be less restrictive */
struct stack_flush {
  struct stack_data *stack;
  uval data2;   /* must be 0 */
/*  uval data3; */  /* must be 0 */
};

/* This global variable should be in it's `permanent' position when */
/* this call occurs */
struct global_flush {
  ref variable;
  uval offset;
/* uval data3; */    /* must be 0 */
};
#endif

typedef struct arg_t {
  arg_type type;
  union {
    struct generic        gn;
    struct stack_reg      sr;
    struct global_reg     gr;
    struct random_reg     rr;
    struct preserved_reg  pr;
  }arg;
}arg_t;

typedef struct arg_array {
  arg_t a[MAX_REG];
}arg_array;

typedef struct function_info {
  func_ptr_t func;
  arg_array arg;  
  arg_array ret;  
}func_t;

/* ( P: (func_t * info) -- ) */
void execute_function(void);

#define VV_def  { \
  /* eax */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* ecx */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* edx */ { RANDOM_REG , {{ 0, 0 }}}, \
  /* ebx */ { PRESERVED_REG , {{ 0, 0 }}}, \
  /* esp */ { GLOBAL_REG , {{ &rstack, offsetof(stack_data,place_pushed) }}}, \
  /* ebp */ { PRESERVED_REG , {{ 0, 0 }}}, \
  /* esi */ { PRESERVED_REG , {{ 0, 0 }}}, \
  /* edi */ { PRESERVED_REG , {{ 0, 0 }}} \
	  } 
	    

extern arg_array VV;

#endif /* FUNCTION_H */
