#include "throf.h"
#include "dictionary.h"
#include "interpret.h"
#include "stack.h"
#include "io.h"
#include "operators.h"
#include "error.h"
#include "colon-compiler.h"

unsigned state = INTERPRETING;  /* true when interpreting */

void VVstate(void)
{
  ppush(&state);
}

/* this isn't ANSForth find */
void VVfind(void)
{
  /* 
   ( P: c-addr -- c-addr | xt | xt ) ( D: u -- u 0 | 1 | -1 )
   \ takes a counted string with length
  */
  entry * e;
  unsigned length;
  char * string;
  string = ppop();
  length = dpop();
  e = find_func(string,length);
  if (e == NIL)
    {
      ppush(string);
      dpush(length);
      dpush(0);
    }
  else
    {
      ppush(e->execution);
      dpush((e->flags & IMM)?1:-1);
    }
}

void VVinterpret_word(void)
{
  int flag;
  VVbl();
  VVparse_word();
  VVfind();
  flag = dpop();
  if (state == INTERPRETING)         /* I'm interpreting */
    {
      if (flag)
	execute_function();
      else
	{
	  VVdup();
	  if (dpop()!= 0)
	    {
	      VVatoi();
	    }
	  else
	    {
	      dpop();
	      ppop();
	    }
	}
    }
  else              /* I'm compiling */
    {
      if (flag == 1) /* this is an immediate word, just execute it */
	{
	  execute_function();
	}
      else if (flag == -1)  /* this is a word */
	{
	  VVcompile_call();
	}
      else  /* this is a constant add it to the word */
	{
	  VVdup();
	  if (dpop() != 0)
	    {
	      VVatoi();
	      VVdconst();
	    }
	  else
	    {
	      dpop();
	      ppop();
	    }
	}
    }
}

void VVinterpret(void)
{
  char ok_str[] = "ok> ";
  while(1)
    {
      if (eol())
	{
	  dpush(sizeof(ok_str) -1);
	  ppush(ok_str);
	  VVtype();
	  VVquery();  /* I really should use refill here :( */
	}
      VVinterpret_word();
    }
}


#if 0
entry * last_word = 0;

void VVcolon(void)
{
  VVbl();
  VVparse_word();
  VVcolon_noname();

  /* this needs to happen after semicolon!! */
  func = ppop();
  name = ppop();
  name_len = dpop();
  last_word = register_func(name, name_len, func, 0);
}

void VVimmediate(void)
{
  if (last_word != 0)
    last_word->flags |= IMM;
}
#endif 

C( D: u -- )  C( P: c-addr xt -- rw) /* an rw is a reference to a word */
void VVregister_word(void)
{
  char * name;
  unsigned name_len;
  func_t *func;
  entry * registered_word;

  
  func = ppop();
  name = ppop();
  name_len = dpop();
  registered_word = register_func(name, name_len, func, 0);
  ppush(registered_word);
}

C( P: rw -- )
void VVmake_immediate(void)
{
  entry * word;
  word = ppop();
  word->flags |= IMM;
}

void VVleft_bracket(void)
{
  state = INTERPRETING;  /* switch to compilation state */
}

void VVright_bracket(void)
{
  state = COMPILING;     /* switch back to compiling */
}

void VVpostpone(void)
{
  /* I should probably have a bit of code here that changes 
     the value of state, but I won't, I'll just call it.
     Another ANS incompatibility 
   */
  VVbl();
  VVparse_word();

  VVfind();
  if (dpop())
    VVcompile_call();
  else  /* this is a constant add it to the word */
    {
      VVatoi();
      VVdconst();
    }
}


/*------------------ register the functions ---------------- */
void register_interpret(void)
{
#define S(X) X , (sizeof(X) -1)

#define RET0 &VV
#define CALLs0 &VV

  register_function(S("interpret_word"), FUNC_ADDR(VVinterpret),       0,   CALLs0, RET0);
  register_function(S("interpret"),      FUNC_ADDR(VVinterpret),       0,   CALLs0, RET0);
  register_function(S("state"),          FUNC_ADDR(VVstate),           0,   CALLs0, RET0);
  register_function(S("find"),           FUNC_ADDR(VVfind),            0,   CALLs0, RET0);
  register_function(S("["),              FUNC_ADDR(VVleft_bracket),    IMM, CALLs0, RET0);
  register_function(S("]"),              FUNC_ADDR(VVright_bracket),   0,   CALLs0, RET0);
  register_function(S("register_word"),  FUNC_ADDR(VVregister_word),   0,   CALLs0, RET0);
  register_function(S("make_immediate"), FUNC_ADDR(VVmake_immediate),  0,   CALLs0, RET0);
  register_function(S("postpone"),       FUNC_ADDR(VVpostpone),        IMM, CALLs0, RET0);
  

#undef S
#undef CALLs0
#undef RET0
}

