#include <stdio.h>
#include <ctype.h>
#include "throf.h"
#include "io.h"
#include "memory.h"
#include "stack.h"
#include "dictionary.h"

/* remember to position in just past the last character read */

char * input_buffer = 0;
unsigned input_buffer_length = 0;
unsigned input_buffer_space = 0;
FILE * io_source = stdin;
FILE * io_sink = stdout;

char * input_source;
unsigned input_length;
unsigned source_id = 0;  /* 0 user input device, -1 string */
unsigned in = 0;
void VVsource_id(void)
{
  dpush(source_id);
}



#define INPUT_CHUNK_SIZE 80
#define EOL '\n'
void VVrefill(void)
{
  int c;
  
  if (source_id == -1)
    {
      dpush(FALSE);
      return;
    }

  if (input_buffer == 0)
    {
      input_buffer = allocate(INPUT_CHUNK_SIZE);
      input_buffer_space = INPUT_CHUNK_SIZE;
    }
  input_buffer_length = 0;
  for(c = getc(io_source); (c != EOL) && (c != EOF); input_buffer_length++)
    {
      if (input_buffer_length == input_buffer_space)
	{
	  input_buffer = reallocate(input_buffer_space, 
				    input_buffer_space+INPUT_CHUNK_SIZE, input_buffer);
	  input_buffer_space += INPUT_CHUNK_SIZE;
	}
      input_buffer[input_buffer_length] = c;
      c = getc(io_source);
    }
  input_source = input_buffer;
  input_length = input_buffer_length;
  dpush(TRUE);
}

int eol(void)  { return in >= input_length; }

void VVquery(void)
{
  /* ( -- )
   */
  source_id = 0;
  in = 0;
  VVrefill();
  dpop();  /* remove the flag! */
}

void VVtype(void)
{
  /* ( P: c-addr -- ) ( D: u -- )
     \ if u > 0 display the character string specified by c-addr and u
   */
  unsigned i;
  char * c;
  c = ppop();
  i = dpop();
  while(i != 0)
    {
      putc(*c,io_sink);
      i--;
      c++;
    }
  fflush(io_sink);
}

void VVsource(void)
{
  /* ( P: -- c-addr) ( D: -- u)
     \ returns the length and addres of the input buffer 
  */
  ppush(input_source);
  dpush(input_length);
}

void VV_in(void)
{
  /*
  ( P: -- a-addr)
  */
  ppush(&in);
}

void VVeol(void)
{ 
  dpush(eol()?TRUE:FALSE); 
}

void VVparse(void)
{
  /*
    ( D: <char> -- u) (P: -- c-addr)
  */
  char delimiter;
  char * start;
  unsigned length;

  delimiter = dpop();
  start = &input_source[in];
  length = 0;

  while( !eol() && ( delimiter != input_source[in]))
    {
      in++;
      length++;
    }
  if (input_source[in] == delimiter)
    in++;
  ppush(start);
  dpush(length);
}

int test(int delimiter, int character)
{
#if 0
  return (delimiter == ' ')?isspace(character):delimiter == character;
#else
  return delimiter == character;
#endif
}

void VVparse_word(void)
{
  /*  "<chars>ccc<char>"
    ( D: <char> -- u) (P: -- c-addr)
  */
  char delimiter;
  char * start;
  unsigned length;

  delimiter = dpop();

  while( !eol() && test(delimiter,input_source[in]) )
    in++;

  start = &input_source[in];
  length = 0;

  while( !eol() && !test(delimiter,input_source[in]))
    {
      in++;
      length++;
    }
  if (test(delimiter,input_source[in]))
    in++;
  ppush(start);
  dpush(length);
}

void VVbl(void)
{
  dpush(' ');
}

void VVdot(void)  /* this could well be an inferior implementation */
{
  sval n;
  n = dpop();
  fprintf(io_sink, "%d", n);
}

void VVcr(void)
{
  putc('\n',io_sink);
}

C( u -- ) C( P: c-addr -- file )
void VVopen(void)
{
  FILE *file;
  uval length;
  char *string;

  length = dpop(); string = ppop();
  /* I shouldn't be destructive like this */
  string[length] = '\0';
  file = fopen(string,"r");
  ppush(file);
}

C( P: file -- )
void VVclose(void)
{
  FILE * file;
  file = ppop();
  fclose(file);
}

C( P: -- a-addr )
void VVio_source(void)
{
  ppush(&io_source);
}

C( P: -- a-addr )
void VVio_sink(void)
{
  ppush(&io_sink);
}

C( P: -- file )
void VVstdin(void) {  ppush(stdin); }
void VVstdout(void) { ppush(stdout); }
void VVstderr(void) { ppush(stderr); }

/*------------------ register the functions ---------------- */
void register_io(void)
{
#define S(X) X , (sizeof(X) -1)

#define RET0 &VV
#define CALLs0 &VV

  register_function(S("source-id"), FUNC_ADDR(VVsource_id),      0, CALLs0, RET0);
  register_function(S("source"),    FUNC_ADDR(VVsource),         0, CALLs0, RET0);
  register_function(S("refill"),    FUNC_ADDR(VVrefill),         0, CALLs0, RET0);
  register_function(S("query"),     FUNC_ADDR(VVquery),          0, CALLs0, RET0);
  register_function(S("type"),      FUNC_ADDR(VVtype),           0, CALLs0, RET0);
  register_function(S(">in"),       FUNC_ADDR(VV_in),            0, CALLs0, RET0);
  register_function(S("parse"),     FUNC_ADDR(VVparse),          0, CALLs0, RET0);
  register_function(S("word"),      FUNC_ADDR(VVparse_word),     0, CALLs0, RET0);
  register_function(S("bl"),        FUNC_ADDR(VVbl),             0, CALLs0, RET0);
  register_function(S("eol"),       FUNC_ADDR(VVeol),            0, CALLs0, RET0);
  register_function(S("."),         FUNC_ADDR(VVdot),            0, CALLs0, RET0);
  register_function(S("cr"),        FUNC_ADDR(VVcr),             0, CALLs0, RET0);
  register_function(S("io_source"), FUNC_ADDR(VVio_source),      0, CALLs0, RET0);
  register_function(S("io_sink"),   FUNC_ADDR(VVio_source),      0, CALLs0, RET0);
  register_function(S("open"),      FUNC_ADDR(VVopen),           0, CALLs0, RET0);
  register_function(S("close"),     FUNC_ADDR(VVclose),          0, CALLs0, RET0);
  register_function(S("stdin"),     FUNC_ADDR(VVstdin),          0, CALLs0, RET0);
  register_function(S("stdout"),    FUNC_ADDR(VVstdout),         0, CALLs0, RET0);
  register_function(S("stderr"),    FUNC_ADDR(VVstderr),         0, CALLs0, RET0);

#undef S
#undef CALLs0
#undef RET0
}


