#include <stdlib.h>
#include <stdio.h>
#include "memory.h"

/* note the first word of a memory block is it's allocated size,
   it may be a few bytes less then it's actual size!
 */

#if 1
static void * last;
static unsigned last_size;

#define MEM_SIZE (1024*1024)

static char buffer[MEM_SIZE];
static unsigned place = 0;

static void * get_mem(unsigned size)
{
  void * space;
  if ((place + size) >= MEM_SIZE)
    {
      fprintf(stderr,"sorry memory stack has been exhausted, aborting!!\n");
      exit(EXIT_FAILURE);
    }
  space = &buffer[place];
  place += size;
  return space;
}

static void free_mem(unsigned size) /* release the last size bytes */
{
  if ((place - size) <0)
    {
      fprintf(stderr,"sorry memory stack has been depleted, aborting!!\n");
      exit(EXIT_FAILURE);
    }
  place -= size;
}

static unsigned round(unsigned size)
{
  return (size + 7) & 0xFFFFFFF8; /* round to nearest multiple of 8 */
}

void * allocate(unsigned size)
{
  unsigned *place;
  last_size = size;
  place = get_mem(round(size));
  last = place;
  return last;
}

void grow(int amount, void * place)
{
  void * spot;
  if (place != last)
    {
      fprintf(stderr,"Attempt to grow a block that wasn't the least recently allocated!!\n");
      exit(EXIT_FAILURE);
    }
  free_mem(round(last_size));
  last_size += amount;
  spot = get_mem(round(last_size));
  if (spot != last)
    {
      fprintf(stderr,"Eric messed up when he rote grow, fix it, aborting!!\n");
      exit(EXIT_FAILURE);
    }
  return;
}

void *reallocate(unsigned old_size, unsigned new_size, void * old)
{
  if (old == last)
    {
      grow(new_size - old_size,old);
    }
  else if (new_size > old_size)
    {
      char * new_p, *old_p;
      unsigned i;
      old_p = old;
      new_p = allocate(new_size);
      for(i = 0; i < old_size; i++)
	new_p[i] = old_p[i];
      return (void *) new_p;
    }
  else if (round(old_size) == round(new_size))
    {
    }
  else /* if (new_size < old_size) */
    {
    }

  return old;
}

#else
void * allocate(unsigned size)
{
  void * ptr = malloc(size);
  if ( ptr == 0)
    {
      fprintf(stderr,"couldn't allocate memory, aborting!!!\n");
      exit(EXIT_FAILURE);
    }
  return ptr;
}
#endif

#if 0
void * reallocate(unsigned old_size, unsigned new_size, void * last)
{
  char * p;
  if (old_size < new_size)
    {
      unsigned i;
      p = allocate(new_size);
      for(i = 0; i < new_size; i++)
	p[i] = *((char *)(last));
    }
  else
    p = last;
  return (void *)p;
}
#endif
