#include <stdlib.h>
#include "throf.h"
#include "operators.h"
#include "function.h"
#include "stack.h"
#include "dictionary.h"
#include "spaghetti-stack.h"
#include "io.h"
#include "interpret.h"
#include "error.h"
#include "memory.h"

C( D: u -- u u )
void VVdup(void)
{
  dpush(dpeek(0));
}

C( D: u -- i ) C( P: c-addr -- )
void VVatoi(void)
{
  char *buffer;
  char *string;
  unsigned len;
  unsigned i;
  len = dpop();
  string = ppop();
  buffer = alloca(len +1);

  for(i = 0; i < len; i++)
      buffer[i] = string[i];
  buffer[len] ='\0';

  dpush(atoi(buffer));
}

void VVadd(void)
{
  uval a, b;
  a = dpop();
  b = dpop();
  dpush(a+b);
}

void VVsub(void)
{
  uval a, b;
  a = dpop();
  b = dpop();
  dpush(b-a);
}

void VVbye(void)
{
  exit(EXIT_SUCCESS);
}

void VVslash(void)
{
  sval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush(n1/n2);
}

void VVmod(void)
{
  sval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush(n1%n2);
}

void VVstar(void)
{
  uval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush(n1*n2);
}

void VVshow_stack(void)
{
  uval i;
  i = ddepth();
  while(i != 0)
    {
      dpush(dpeek( i -1));
      VVdot();
      VVcr();
      i--;
    };
}

void VVdot_p(void)
{
  uval i;
  i = pdepth();
  while(i != 0)
    {
      dpush((uval)ppeek(i -1));
      VVdot();
      VVcr();
      i--;
    };
}

void VVdot_r(void)
{
  uval i;
  sync_return_stack();
  i = rdepth();
  while(i != 0)
    {
      dpush(rpeek(i -1));
      VVdot();
      VVcr();
      i--;
    };
}

void VVstar_slash_mod(void)
{
  typedef long long pole;
  pole d;
  sval n1,n2,n3;
  n3 = dpop();
  n2 = dpop();
  n1 = dpop();
  
  d = ((pole)n1) * ((pole)n2);
  dpush(d/n3);
  dpush(d%n3);  
}

void VVswap(void)
{
  uval n1,n2;
  n2 = dpop();
  n1 = dpop();
  dpush(n2);
  dpush(n1);
}

void VVdrop(void)
{
  uval n1;
  n1 = dpop();
}

C( P: a-addr -- ) C( D: -- u)
void VVat(void)
{
  unsigned * a;
  a = ppop();
  dpush(*a);
}

C( P: a-addr -- )  C( D: x -- ) 
void VVbang(void)
{
  unsigned *a;
  uval value;

  value = dpop();
  a = ppop();
  *a = value;
}

C( D: -- false )
void VVfalse(void)
{
  dpush(FALSE);
}

C( D: -- true )
void VVtrue(void)
{
  dpush(TRUE);
}

void VVtick(void)
{
  unsigned flag;
  VVbl();
  VVparse_word();
  VVfind();
  flag = dpop();
  if(flag == 0)
    {
      char string[] = "Hey you tried to tick a non name, that ticks me off, aborting!!\n";
      ppush(string);
      dpush(sizeof(string) -1);
      VVfatal_error();
    }
}

C( u --  ) C( P: -- p )
void VValloc(void)  /*  this isn't ans forth, but it's the primitive I have */
{
  int amount;
  void * block;
  
  amount = dpop();
  block = allocate(amount);

  ppush(block);
}

C( p -- ) C( P: -- p )
void VVdtop(void)
{
  ppush((void *)dpop());
}

C( -- p ) C( P: p -- )
void VVptod(void)
{
  dpush((uval)ppop());
}

C( Xu Xu-1 ... Xo u -- Xu-1 .. Xo Xu)
void VVroll(void)
{
  uval i, u, Xu;
  u = dpop();
  for (i = 0; i < u; i++)
    ppush((ref)dpop());
  
  Xu = dpop();
  
  for(i = 0; i < u; i++)
    dpush((uval)ppop());
  
  dpush(Xu);
}

C( Xu Xu-1 ... Xo u -- Xu Xu-1 ...Xo Xu )
void VVpick(void)
{
  uval u;
  u = dpop();
  dpush(dpeek(u));
}

C( D: u -- ) C( P: Xu Xu-1 ... Xo -- Xu-1 .. Xo Xu)
void VVproll(void)
{
  uval i, u;
  ref Xu;
  u = dpop();
  for (i = 0; i < u; i++)
    dpush((uval)ppop());
  
  Xu = ppop();
  
  for(i = 0; i < u; i++)
    ppush((ref)(dpop()));
  
  ppush(Xu);
}

C( D: u -- ) C( P: Xu Xu-1 ... Xo -- Xu Xu-1 ...Xo Xu )
void VVppick(void)
{
  uval u;
  u = dpop();
  ppush(ppeek(u));
}

C( P: p -- )
void VVpdrop(void)
{
  ppop();
}

C( u -- ) C( P: c-addr -- c-addr)  /* returns address of a counted string */
void VVsave_string(void)
{
  uval length, i;
  char *string, *counted_string;

  length = dpop();
  string = ppop();
  dpush(length + sizeof(uval));
  VValloc();
  counted_string = ppop();
  ppush(counted_string);
  dpush(length);
  VVbang();
  for(i = 0; i < length; i++)
    {
      counted_string[i+ sizeof(uval)] = string[i];
    }
  ppush(counted_string);
}

C( u1 u2 -- flag)
void VVequal(void)
{
  uval u1,u2;
  u2 = dpop();
  u1 = dpop();
  dpush( (u1 == u2)? TRUE :FALSE);
}

C( X1 X2 -- x1&x2 )
void VVand(void)
{
  uval X1, X2;
  X2 = dpop();
  X1 = dpop();
  dpush(X1 & X2);
}

C( X1 X2 -- x1|x2 )
void VVor(void)
{
  uval X1, X2;
  X2 = dpop();
  X1 = dpop();
  dpush(X1 | X2);
}

C( X2 X1 -- X1^X2 )
void VVxor(void)
{
  uval X1, X2;
  X1 = dpop();
  X2 = dpop();
  dpush(X1^X2);
}

/*----------------------  Start registration ------------------------------*/

void register_operators(void)
{
  /* quick synopsis of i386 calling conventions for C */
 
#define S(X) X , (sizeof(X) -1)

#define RET0 &VV
#define CALLs0 &VV

  register_function(S("bye"),    FUNC_ADDR(VVbye),            0, CALLs0, RET0);
  register_function(S(".S"),     FUNC_ADDR(VVshow_stack),     0, CALLs0, RET0);
  register_function(S(".P"),     FUNC_ADDR(VVdot_p),          0, CALLs0, RET0);
  register_function(S(".R"),     FUNC_ADDR(VVdot_r),          0, CALLs0, RET0);
  register_function(S("+"),      FUNC_ADDR(VVadd),            0, CALLs0, RET0);
  register_function(S("-"),      FUNC_ADDR(VVsub),            0, CALLs0, RET0);
  register_function(S("/"),      FUNC_ADDR(VVslash),          0, CALLs0, RET0);
  register_function(S("mod"),    FUNC_ADDR(VVmod),            0, CALLs0, RET0);
  register_function(S("*"),      FUNC_ADDR(VVstar),           0, CALLs0, RET0);
  register_function(S("swap"),   FUNC_ADDR(VVswap),           0, CALLs0, RET0);
  register_function(S("*/mod"),  FUNC_ADDR(VVstar_slash_mod), 0, CALLs0, RET0);
  register_function(S("drop"),   FUNC_ADDR(VVdrop),           0, CALLs0, RET0);
  register_function(S("@"),      FUNC_ADDR(VVat),             0, CALLs0, RET0);
  register_function(S("!"),      FUNC_ADDR(VVbang),           0, CALLs0, RET0);
  register_function(S("true"),   FUNC_ADDR(VVtrue),           0, CALLs0, RET0);
  register_function(S("false"),  FUNC_ADDR(VVfalse),          0, CALLs0, RET0);
  register_function(S("execute"),FUNC_ADDR(execute_function), 0, CALLs0, RET0);
  register_function(S("'"),      FUNC_ADDR(VVtick),           0, CALLs0, RET0);
  register_function(S("atoi"),   FUNC_ADDR(VVatoi),           0, CALLs0, RET0);
  register_function(S("dup"),    FUNC_ADDR(VVdup),            0, CALLs0, RET0);
  register_function(S("alloc"),  FUNC_ADDR(VValloc),          0, CALLs0, RET0);
  register_function(S("P>D"),    FUNC_ADDR(VVptod),           0, CALLs0, RET0);
  register_function(S("D>P"),    FUNC_ADDR(VVdtop),           0, CALLs0, RET0);
  register_function(S("roll"),   FUNC_ADDR(VVroll),           0, CALLs0, RET0);
  register_function(S("pick"),   FUNC_ADDR(VVpick),           0, CALLs0, RET0);
  register_function(S("save_string"),     FUNC_ADDR(VVsave_string),   0, CALLs0, RET0);
  register_function(S("proll"),  FUNC_ADDR(VVproll),          0, CALLs0, RET0);
  register_function(S("ppick"),  FUNC_ADDR(VVppick),          0, CALLs0, RET0);
  register_function(S("pdrop"),  FUNC_ADDR(VVpdrop),          0, CALLs0, RET0);
  register_function(S("="),      FUNC_ADDR(VVequal),          0, CALLs0, RET0);
  register_function(S("and"),    FUNC_ADDR(VVand),            0, CALLs0, RET0);
  register_function(S("or"),     FUNC_ADDR(VVor),             0, CALLs0, RET0);
  register_function(S("xor"),    FUNC_ADDR(VVxor),            0, CALLs0, RET0);
  register_function(S("fatal_error"), FUNC_ADDR(VVfatal_error), 0, CALLs0, RET0);

#undef S
#undef CALLs0
#undef RET0
}




