#include <stdio.h>
#include <stdlib.h>

#include "throf.h"
#include "stack.h"

#define DSTACK_SIZE 200
#define PSTACK_SIZE 200
#define FSTACK_SIZE 20
#define RSTACK_SIZE 200000
#define RSTACK_START rstack_space

static cell dstack_space[DSTACK_SIZE];
static cell pstack_space[PSTACK_SIZE];
static cell fstack_space[FSTACK_SIZE];
#if 0
static cell rstack_space[RSTACK_SIZE];
#endif

stack_data dstack, pstack, fstack, rstack;

/*----------------------- stack initation -----------------------*/

#define RSTACK_BOT rstack.place_pushed
#define RSTACK_TOP (RSTACK_BOT + (RSTACK_DIR * (RSTACK_SIZE -1)))

void init_stacks(void (*the_rest)(void))
{
  cell * words;
  stack_init("data",   dstack_space,DSTACK_SIZE,up,&dstack);
  stack_init("pointer",pstack_space,PSTACK_SIZE,up,&pstack);
  stack_init("float",  fstack_space,FSTACK_SIZE,up,&fstack);
  
  /* now I must inline return stack initialization */
  rstack.direction = RSTACK_DIR;
  rstack.name = "return";
  words = alloca(sizeof(cell) * FRAME_DEPTH);   /* allocate stack space for the frame */
  sync_return_stack();
  rstack.frame =  rstack.place_pushed + RSTACK_DIR;
  rstack.bottom = rstack.frame        - (RSTACK_DIR * FRAME_DEPTH);
  rstack.top =    rstack.bottom +       (RSTACK_DIR * (RSTACK_SIZE -1));
  frame_top_set(rstack.top, rstack.frame, &rstack);
  frame_previous_set(0, rstack.frame, &rstack);
  the_rest();
}

#if DISABLE_STACK_CONTINUATIONS 
void dis()
{
  dstack.place_to_pop = dstack.place_pushed;
  pstack.place_to_pop = pstack.place_pushed;
  fstack.place_to_pop = fstack.place_pushed;
  rstack.place_to_pop = rstack.place_pushed;  /* like this is accurate :) */
}
#define DIS dis();
#else
#define DIS
#endif
/*------------- data stack ------------*/

void dpush(uval val)
{
  DIS
  push(val,&dstack);
}

uval dpop(void)
{
  DIS
  return pop(&dstack);
}

uval ddepth(void)
{
  DIS
  return depth(&dstack);
}

uval dpeek(uval n)
{
  DIS
  return peek(n,&dstack);
}

/* this is a really lame definition */
/* I should need to involve other stacks etc. */
void droll(uval u)
{
  uval i, Xu;
  DIS
  for(i = 0; i < u; i++)
    ppush((ref)dpop());
  
  Xu = dpop();
    
  for (i = 0; i < u; i++)
    dpush((uval)ppop());

  dpush(Xu);
}

/*--------- pointer stack -------------*/

void ppush(ref val)
{
  DIS
  push(PTOC(val),&pstack);
}

ref ppop(void)
{
  DIS
  return CTOP(pop(&pstack));
}

uval pdepth(void)
{
  DIS
  return depth(&pstack);
}

ref ppeek(uval n)
{
  DIS
  return CTOP(peek(n,&pstack));
}

/* this is a really lame definition */
/* I should need to involve other stacks etc. */
void proll(uval u)
{
  uval i; ref  Xu;
  DIS
  for(i = 0; i < u; i++)
    dpush((uval)ppop());
  
  Xu = ppop();
    
  for (i = 0; i < u; i++)
    ppush((ref)dpop());

  ppush(Xu);
}

/*--------- float stack ---------*/

#define CELLS_TO_THE_FLOAT (((sizeof(float) +sizeof(cell) -1)/sizeof(cell)) * sizeof(cell))

typedef union floatcell {
  real r;
  cell c[CELLS_TO_THE_FLOAT];
}fcell;

void fpush(real val)
{
  fcell f;
  unsigned i;
  DIS
  f.r = val;
  for(i = 0; i < CELLS_TO_THE_FLOAT; i++)
    push(f.c[i],&fstack);
}

real fpop(void)
{
  fcell f;
  unsigned i;
  DIS
  for(i = 0; i < CELLS_TO_THE_FLOAT; i++)
    f.c[CELLS_TO_THE_FLOAT - i - 1] = pop(&fstack);
  return f.r;
}

uval fdepth(void)
{
  DIS
  return depth(&fstack)/ CELLS_TO_THE_FLOAT;
}

real fpeek(uval n)
{
  fcell f;
  unsigned i;
  DIS
  for(i = 0; i < CELLS_TO_THE_FLOAT; i++)
    f.c[i] = peek(n*CELLS_TO_THE_FLOAT + i,&fstack);
  return f.r;
}


/*------------- return stack ------------*/

void rpush(uval val)
{
  DIS
  push(val,&rstack);
}

uval rpop(void)
{
  DIS
  return pop(&rstack);
}

uval rdepth(void)
{
  DIS
  return depth(&rstack);
}

uval rpeek(uval n)
{
  DIS
  return peek(n,&rstack);
}

